/*---------------------------------------------------------------------------
 * Copyright (C) 2008-2009 - Emanuele Bovisio
 *
 * This file is part of Mulk.
 *
 * Mulk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Mulk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mulk.  If not, see <http://www.gnu.org/licenses/>.
 *---------------------------------------------------------------------------*/

#include <unistd.h>
#include "defines.h"
#include "buffer_array.h"
#include "url_list.h"
#include "parse.h"
#include "jpg_obj.h"
#include "file_obj.h"

#define NUM_OPTIONS ((int) (sizeof(options) / sizeof(option_t)))

typedef enum type_option_t {
	OPTION_BOOL,
	OPTION_INT,
	OPTION_STRING,
	OPTION_EXIT,
	OPTION_VERSION,
	OPTION_TITLE
} type_option_t;

option_value_t option_values = {
	0,	/* int quiet; */
	0,	/* int verbosity; */
	1, 	/* int depth; */
	NULL, /* char *option_filename; */
	NULL, /* char *url_filename; */
#ifdef ENABLE_METALINK
	NULL, /* char *metalink_filename; */
	NULL, /* char *metalink_location; */
	NULL, /* char *metalink_os; */
#endif
	NULL, /* char *report_filename; */
	NULL, /* char *report_csv_filename; */
	2,	/* int max_sim_conns_per_host; */
	50,	/* int max_sim_conns; */
	0, /* int disable_save_tree; */
	0, /* int save_gif_image; */
	0, /* int save_png_image; */
	0, /* int save_jpeg_image; */
	0, /* int min_image_width; */
	0, /* int max_image_width; */
	0, /* int min_image_height; */
	0, /* int max_image_height; */
	NULL, /* char *output_directory; */
};

typedef struct option_t {
	char *short_opt;
	char *long_opt;
	char *description;
	void *value;
	type_option_t type;
	int min_val;
	int max_val;
	char *error;
} option_t;

char version_number[] = VERSION;

option_t options[] = {
	{"", "--version", "display version number and exit", NULL, OPTION_VERSION, 0, 0, NULL},
	{"-h", "--help", "display this help and exit", NULL, OPTION_EXIT, 0, 0, NULL},
	{"-q", "--quiet", "quiet (no output)", &option_values.quiet, OPTION_BOOL, 0, 0, NULL},
	{"-v", "--verbose", "verbosity", &option_values.verbosity, OPTION_BOOL, 0, 0, NULL},
	{"-d", "--depth", "depth", &option_values.depth, OPTION_INT, 1, 10, "wrong depth values"},
	{"-t", "--option-file", "text file with list of options", &option_values.option_filename, OPTION_STRING, 0, 0, NULL},
	{"-f", "--url-file", "text file with list of URLs to download", &option_values.url_filename, OPTION_STRING, 0, 0, NULL},
#ifdef ENABLE_METALINK
	{"-l", "--metalink-file", "metalink filename", &option_values.metalink_filename, OPTION_STRING, 0, 0, NULL},
	{"", "--metalink-location", "locations to use when downloading metalink files", &option_values.metalink_location, OPTION_STRING, 0, 0, NULL},
	{"", "--metalink-os", "operating system to download in metalink files", &option_values.metalink_os, OPTION_STRING, 0, 0, NULL},
#endif /* ENABLE_METALINK */
	{"-r", "--report-file", "report filename", &option_values.report_filename, OPTION_STRING, 0, 0, NULL},
	{"-c", "--report-csv-file", "report CSV filename", &option_values.report_csv_filename, OPTION_STRING, 0, 0, NULL},
	{"-p", "--max-sim-conns-per-host", "max simultaneous connections per host", 
		&option_values.max_sim_conns_per_host, OPTION_INT, 1, 5, 
		"wrong max simultaneous connections per host value"},
	{"-m", "--max-sim-conns", "max simultaneous connections", &option_values.max_sim_conns, OPTION_INT,
		1, 50, "wrong max simultaneous connections value"},
	{"-x", "--disable-site-save", "don't save whole site tree to disk (enabled by default)", &option_values.disable_save_tree, OPTION_BOOL, 0, 0, NULL},
	{"-g", "--save-gif-image", "save GIF images to output directory", &option_values.save_gif_image, OPTION_BOOL, 0, 0, NULL},
	{"-n", "--save-png-image", "save PNG images to output directory", &option_values.save_png_image, OPTION_BOOL, 0, 0, NULL},
	{"-j", "--save-jpeg-image", "save JPEG images to output directory", &option_values.save_jpeg_image, OPTION_BOOL, 0, 0, NULL},
	{"", "", "Image Filtering (active only with --save-gif-image, --save-png-image or --save-jpeg-image)", NULL, OPTION_TITLE, 0, 0, NULL},
	{"", "--min-image-width", "minimum image width", &option_values.min_image_width, OPTION_INT,
		0, INT_MAX, "wrong minimum image width"},
	{"", "--max-image-width", "maximum image width", &option_values.max_image_width, OPTION_INT,
		0, INT_MAX, "wrong maximum image width"},
	{"", "--min-image-height", "minimum image height", &option_values.min_image_height, OPTION_INT,
		0, INT_MAX, "wrong minimum image height"},
	{"", "--max-image-height","maximum image height", &option_values.max_image_height, OPTION_INT,
		0, INT_MAX, "wrong maximum image height"},
	{"", "--output-dir", "output directory", &option_values.output_directory, OPTION_STRING, 0, 0, NULL},
};

int is_printf(log_level)
{
	return (option_values.verbosity || (!option_values.quiet && (log_level) < MINFO));
}

static void printf_usage(void)
{
	int i;
	printf("Mulk %s, a non-interactive multi-connection network downloader with image filtering"
#ifdef ENABLE_METALINK
		" and Metalink support"
#endif
		".\n", version_number);
	printf("Usage: Mulk [OPTION]... URL...\n\n");

	printf("Options:\n");
	for (i = 0; i < NUM_OPTIONS; i++)
		if (options[i].type == OPTION_TITLE)
			printf("\n      %s\n\n", options[i].description);
		else
			printf("  %2s%c %-25s%s\n", options[i].short_opt, *(options[i].short_opt) ? ',' : ' ', 
				options[i].long_opt, options[i].description);

	printf("\n");
}

int set_option(char *option, char *value, int remove_minus, int *value_used)
{
	int j, val;

	*value_used = 0;

	for (j = 0; j < NUM_OPTIONS; j++)
		if ((!remove_minus && (!strcmp(options[j].short_opt, option)
			|| !strcmp(options[j].long_opt, option)))
			|| (remove_minus && *(options[j].long_opt) && !strcmp(options[j].long_opt+2, option)))
			break;

	if (j >= NUM_OPTIONS) {
		fprintf(stderr, "\nERROR: option not recognised, %s\n\n", option);
		return 1;
	}

	switch (options[j].type) {
		case OPTION_BOOL:
			*((int*) options[j].value) = 1;
			break;
		case OPTION_INT:
			if (!value) {
				fprintf(stderr, "\nERROR: this option (%s) needs to be followed by a value\n\n", option);
				return 1;
			}
			val = atoi(value);
			*value_used = 1;
			if (val < options[j].min_val || val > options[j].max_val) {
				fprintf(stderr, "\nERROR: %s\n\n", options[j].error);
				return 1;
			}
			*((int*) options[j].value) = val;
			break;
		case OPTION_STRING:
			if (!value) {
				fprintf(stderr, "\nERROR: this option (%s) needs to be followed by a value\n\n", option);
				return 1;
			}
			if (*((char**) options[j].value)) {
				fprintf(stderr, "\nERROR: this option (%s) was already specified\n\n", option);
				return 1;
			}
			*((char**) options[j].value) = calloc(strlen(value) + 1, sizeof(char));
			strcpy(*((char**) options[j].value), value);
			*value_used = 1;
			break;
		case OPTION_EXIT:
			return 1;
		case OPTION_VERSION:
			printf("Mulk version %s\n", version_number);
			return 2;
		case OPTION_TITLE:
			break;
		default:
			fprintf(stderr, "\nERROR: option type not valid, %d\n\n", options[j].type);
			return 1;
	}

	return 0;
}

static int set_options(int argc, char **argv)
{
	int i, ret, value_used;
	char *value;

	for (i = 1; i < argc; i++) {
		/* it's an url */
		if (!argv[i] || *argv[i] != '-')
			break;

		if ((i + 1) < argc && argv[i+1] && *argv[i+1] != '-')
			value = argv[i+1];
		else
			value = NULL;

		value_used = 0;
		if ((ret = set_option(argv[i], value, 0, &value_used)))
			return ret;

		if (value_used)
			i++;
	}

	/* read options from a file */
	if (option_values.option_filename) 
		if ((ret = read_option_from_text_file(option_values.option_filename))) 
			return ret;

	if (i >= argc && !option_values.url_filename 
#ifdef ENABLE_METALINK
	  && !option_values.metalink_filename
#endif
	) {
		fprintf(stderr, "\nERROR: url not present\n\n");
		return 1;
	}

	while (i < argc) {
		if (*argv[i] == '-') {
			fprintf(stderr, "\nERROR: options have to come before URLs\n\n");
			return 1;
		}
		if (is_printf(MNOTE))
			printf("Add url to download from command line: %s\n", argv[i]);
		push_unique_url(NULL, argv[i++], 1);
	}

	return 0;
}

static void free_options(void)
{
	int j;

	for (j = 0; j < NUM_OPTIONS; j++)
		if (options[j].type == OPTION_STRING && *((char**) options[j].value))
			free(*((char **) options[j].value));
}

int main(int argc, char **argv)
{
	CURLM *curl_obj;
	CURLMsg *curl_msg;
	long time_to_wait;
	int num_fd, num_msgs, active_handles, opt_ret;
	fd_set read_fd, write_fd, except_fd;
	struct timeval timeout;
	time_t t1, t2, tot_time;
	long download_tot = 0;

	if ((opt_ret = set_options(argc, argv))) {
		if (opt_ret == 1)
			printf_usage();
		return 0;
	}

	(void) time(&t1);

	if (option_values.url_filename)
		read_uri_from_text_file(option_values.url_filename);

#ifdef ENABLE_METALINK
	if (option_values.metalink_filename) {
		if (save_locations(option_values.metalink_location))
			return 1;
		add_new_metalink(option_values.metalink_filename, 1);
	}
#endif /* ENABLE_METALINK */

	reset_buffer_array();
	curl_global_init(CURL_GLOBAL_ALL);

	curl_obj = curl_multi_init();

	curl_multi_setopt(curl_obj, CURLMOPT_MAXCONNECTS, (long)option_values.max_sim_conns);

	for (active_handles = 0; active_handles < option_values.max_sim_conns && init_url(curl_obj);
		active_handles++);

	while (active_handles) {
		while (curl_multi_perform(curl_obj, &active_handles) == CURLM_CALL_MULTI_PERFORM);

		if (active_handles) {
			FD_ZERO(&read_fd);
			FD_ZERO(&write_fd);
			FD_ZERO(&except_fd);

			if (curl_multi_fdset(curl_obj, &read_fd, &write_fd, &except_fd, &num_fd)) {
				if (is_printf(MERR))
					printf("ERROR: fdset\n");
				return 1;
			}

			if (curl_multi_timeout(curl_obj, &time_to_wait)) {
				if (is_printf(MERR))
					printf("ERROR: timeout\n");
				return 1;
			}

			if (time_to_wait < 0)
				time_to_wait = 100;

			if (num_fd == -1)
				sleep(time_to_wait / 1000);
			else {
				timeout.tv_sec = time_to_wait / 1000;
				timeout.tv_usec = (time_to_wait % 1000) * 1000;

				if (0 > select(num_fd+1, &read_fd, &write_fd, &except_fd, &timeout)) {
					if (is_printf(MERR))
						printf("ERROR: num_fd:%d, time_to_wait %li, error message %s (%d)\n",
							num_fd+1, time_to_wait, strerror(errno), errno);
					return 1;
				}
			}
		}

		while ((curl_msg = curl_multi_info_read(curl_obj, &num_msgs))) {
			CURL *e = curl_msg->easy_handle;
			if (curl_msg->msg == CURLMSG_DONE) {
				char *url, *mime_type;
				long resp_code;
				double download_size;

				/* curl_easy_getinfo(curl_msg->easy_handle, CURLINFO_PRIVATE, &url); */
				curl_easy_getinfo(curl_msg->easy_handle, CURLINFO_EFFECTIVE_URL, &url);
				curl_easy_getinfo(curl_msg->easy_handle, CURLINFO_CONTENT_TYPE, &mime_type);
				curl_easy_getinfo(curl_msg->easy_handle, CURLINFO_RESPONSE_CODE, &resp_code);
				curl_easy_getinfo(curl_msg->easy_handle, CURLINFO_SIZE_DOWNLOAD, &download_size);

				download_tot += (long)download_size;

				if (is_printf(MNOTE))
					printf("RESULT: %d (%s), HTTP Code:\"%ld\" Mime-Type:\"%s\" Url:\"%s\" Size:%ld\n",
						curl_msg->data.result, curl_easy_strerror(curl_msg->data.result), 
						resp_code, mime_type ? mime_type : "", url, (long)download_size);

				set_buffer_mime_type(e, mime_type);
				close_buffer(e, url, curl_msg->data.result, resp_code);
				curl_multi_remove_handle(curl_obj, e);
				curl_easy_cleanup(e);
			}
			else {
				close_buffer(e, NULL, CURLE_FAILED_INIT, 0);
				if (is_printf(MERR))
					printf("ERROR: message=%d\n", curl_msg->msg);
			}
		}

		while (active_handles < option_values.max_sim_conns && init_url(curl_obj))
			active_handles++;
	}

	curl_multi_cleanup(curl_obj);
	curl_global_cleanup();

	(void) time(&t2);
	tot_time = t2 - t1;

	if (is_printf(MNOTE)) {
		printf("\nTotal time to download = %ld seconds", tot_time);
		printf("\nTotal size downloaded = %.2f KB\n", (float)download_tot / 1000.0f);
		if (tot_time > 0)
			printf("\nAverage data rate = %.2f KB/s\n", ((float)download_tot) / (tot_time * 1000.0f));
	}

	if (option_values.report_filename)
		print_urls(option_values.report_filename);

	if (option_values.report_csv_filename)
		print_csv_urls(option_values.report_csv_filename);

	free_buffer_array();
	free_urls();
#ifdef ENABLE_METALINK
	free_metalinks();
#endif
	free_options();

	return 0;
}
