/*---------------------------------------------------------------------------
 * Copyright (C) 2008-2009 - Emanuele Bovisio
 *
 * This file is part of Mulk.
 *
 * Mulk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Mulk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mulk.  If not, see <http://www.gnu.org/licenses/>.
 *---------------------------------------------------------------------------*/

#include <string.h>
#include "metalink_list.h"
#include "mime_types.h"
#include "url_list.h"
#include "buffer_array.h"

#define MAX_LOC_OPTION_LENGTH 512
#define LOC_LENGTH 2
#define LOC_OPTION_DELIM ",;.:- "

static char locations[MAX_LOC_OPTION_LENGTH];

metalink_list_t *metalink_top = NULL;
metalink_list_t *metalink_bottom = NULL;


/* Metalink locations */
int save_locations(char *loc_option)
{
	char buf[MAX_LOC_OPTION_LENGTH];
	char *s = buf, *res, *ptr = locations;

	*ptr = 0;
	if (!loc_option || !*loc_option)
		return 0;

	strcpy(buf, loc_option);
	while ((res = strtok(s, LOC_OPTION_DELIM))) {
		if (strlen(res) != LOC_LENGTH) {
			if (is_printf(MERR))
				printf("ERROR: Metalink location length is wrong: %s\n", res);
			return 1;
		}

		if (ptr + LOC_LENGTH + 1 > locations + MAX_LOC_OPTION_LENGTH) {
			if (is_printf(MERR))
				printf("ERROR: too many Metalink locations\n");
			return 1;
		}

		strcpy(ptr, res);
		ptr += LOC_LENGTH;
		if (is_printf(MNOTE))
			printf("Add Metalink location to list: %s\n", res);
		s = NULL;
	}

	return 0;
}

static int is_location_in_list(char *loc)
{
	char *ptr = locations;

	/* if there aren't locations then the test is true */
	if (!*locations)
		return 1;

	if (!loc || strlen(loc) != LOC_LENGTH)
		return 0;

	while (*ptr && strncmplower(ptr, loc, LOC_LENGTH))
		ptr += LOC_LENGTH;

	return *ptr ? 1 : 0;
}


/* chunk_t */

void push_chunk(metalink_file_list_t *file, long start, long length)
{
	chunk_t *elem;

	if (!file)
		return;

	elem = calloc(1, sizeof(chunk_t));
	elem->start = start;
	elem->length = length;
	elem->next = file->chunk;

	file->chunk = elem;
}

void print_chunks(metalink_file_list_t *file)
{
	chunk_t *elem;
	int i = 0;

	if (!file || !is_printf(MINFO))
		return;

	elem = file->chunk;

	while (elem) {
		char *uri_str = elem->used_res ? uri2string(elem->used_res->uri) : "NULL";

		printf("%d: %ld %ld-%ld/%ld, %s\n", ++i, elem->pos, elem->start,
			elem->start+elem->length-1, elem->length, uri_str);

		elem = elem->next;
		if (uri_str)
			free(uri_str);
	} 
}

chunk_t *find_free_chunk(metalink_file_list_t *file)
{
	chunk_t *elem;

	if (!file)
		return NULL;

	elem = file->chunk;

	while (elem) {
		if (!elem->used_res && elem->length > elem->pos)
			return elem;
		elem = elem->next;
	} 
	return NULL;
}

int is_file_downloaded(metalink_file_list_t *file)
{
	chunk_t *elem;

	if (!file)
		return 0;

	if (!(elem = file->chunk))
		return 0;

	while (elem) {
		if (elem->length > elem->pos)
			return 0;
		elem = elem->next;
	} 
	return 1;
}

int file_statistics(metalink_file_list_t *file, int *chunk_completed, int *chunk_total,
	long *byte_downloaded, long *byte_total)
{
	chunk_t *elem;

	*chunk_completed = *chunk_total = 0;
	*byte_downloaded = *byte_total = 0;

	if (!file)
		return 0;

	*chunk_total = file->chunk_number;
	*byte_total = file->size;

	if (!(elem = file->chunk))
		return 0;

	while (elem) {
		if (elem->length == elem->pos) {
			(*chunk_completed)++;
			*byte_downloaded += elem->pos;
		}
		elem = elem->next;
	} 
	return 1;
}

void create_chunks(metalink_file_list_t *file)
{
	int chunk_number = CHUNK_NUMBER, i;
	long last_chunk_size, chunk_size, chunk_remainder, filesize;
	
	if (!file || file->size < 0)
		return;

	filesize = file->size;

	chunk_size = filesize / chunk_number;
	if (chunk_size < MIN_CHUNK_SIZE) {
		chunk_number = filesize / MIN_CHUNK_SIZE;
		if (!chunk_number)
			chunk_number = 1;
		chunk_size = filesize / chunk_number;
	}
	else if (chunk_size > MAX_CHUNK_SIZE) {
		chunk_number = filesize / MAX_CHUNK_SIZE + 1;
		if (!chunk_number)
			chunk_number = 1;
		chunk_size = filesize / chunk_number;
	}

	chunk_remainder = filesize % chunk_number;
	last_chunk_size = chunk_size + chunk_remainder;

#ifdef DEBUG
	printf("%s,%d: filesize:%ld\n", __FUNCTION__, __LINE__, filesize);
	printf("%s,%d: chunk_number:%d\n", __FUNCTION__, __LINE__, chunk_number);
	printf("%s,%d: chunk_size:%ld\n", __FUNCTION__, __LINE__, chunk_size);
	printf("%s,%d: last_chunk_size:%ld\n", __FUNCTION__, __LINE__, last_chunk_size);
#endif /* DEBUG */

	for (i = 0; i < chunk_number && filesize > 0; i++) {
		int size = i ? chunk_size : last_chunk_size;

		filesize -= size;

		push_chunk(file, filesize >= 0 ? filesize : 0, size);
	}
	file->chunk_number = chunk_number;
}

static void free_chunks(metalink_file_list_t *file)
{
	chunk_t *tmp, *elem;

	if (!file)
		return;

	elem = file->chunk;

	while (elem) {
		tmp = elem;
		elem = elem->next;
		free(tmp);
	}
	file->chunk = NULL;
}


/* metalink_resource_list_t */

static void push_metalink_resource(metalink_file_list_t *file, metalink_resource_t *resource)
{
	metalink_resource_list_t *elem;

	if (!file || !resource)
		return;

	elem = calloc(1, sizeof(metalink_resource_list_t));
	elem->resource = resource;
	elem->uri = (UriUriA*)create_absolute_uri(NULL, resource->url);
	elem->prev = file->usable_res_bottom;

	if (file->usable_res_bottom)
		file->usable_res_bottom->next = elem;
	file->usable_res_bottom = elem;

	if (!file->usable_res_top)
		file->usable_res_top = elem;
}

void remove_metalink_resource(metalink_file_list_t *file, metalink_resource_list_t *resource)
{
	metalink_resource_list_t *prev, *next;

	if (!file || !resource || !file->usable_res_top)
		return;

	if (resource->assigned)
	{
		resource->error = 1;
		return;
	}

	prev = resource->prev;
	next = resource->next;

	if (file->usable_res_top == resource)
		file->usable_res_top = next;
	if (file->usable_res_bottom == resource)
		file->usable_res_bottom = prev;

	if (prev)
		prev->next = next;
	if (next)
		next->prev = prev;

	if (resource->uri)
		uri_free(resource->uri);
	free(resource);
}

static int is_valid_resource(metalink_resource_t *resource)
{
	return (!strcmplower(resource->type, USED_TYPE) && is_location_in_list(resource->location));
}

int is_resource_available(metalink_file_list_t *file)
{
	metalink_resource_list_t *elem;

	if (!file)
		return 0;

	elem = file->usable_res_top;
	while (elem) {
		if (!elem->error || elem->assigned)
			return 1;
		elem = elem->next;
	} 

	return 0;
}

static metalink_resource_list_t *find_free_resource(metalink_file_list_t *file)
{
	metalink_resource_list_t *elem = NULL, *max_elem = NULL;
	int max_preference = -1;

	if (!file)
		return NULL;

	elem = file->usable_res_top;
	while (elem) {
		if (!elem->error && elem->resource->preference > max_preference 
				&& is_uri_compatible(elem->uri, elem->resource->maxconnections)) {
			max_preference = elem->resource->preference;
			max_elem = elem;
		}
		elem = elem->next;
	} 

	if (max_elem)
		max_elem->assigned++;

	return max_elem;
}

void free_resources(metalink_file_list_t *file)
{
	metalink_resource_list_t *tmp, *elem;

	if (!file)
		return;

	elem = file->usable_res_top;

	while (elem) {
		tmp = elem;
		elem = elem->next;
		if (tmp->uri)
			uri_free(tmp->uri);
		free(tmp);
	}
	file->usable_res_top = file->usable_res_bottom = NULL;
}


/* metalink_file_list_t */

metalink_file_list_t *create_metalink_file(metalink_file_t *file)
{
	metalink_file_list_t *elem;

	if (!file)
		return NULL;

	elem = calloc(1, sizeof(metalink_file_list_t));
	elem->file = file;
	elem->size = (long) (file->size ? file->size : -1L);

	if (elem->size > 0) 
		create_chunks(elem);

	return elem;
}

UriUriA *find_next_url(metalink_file_list_t *file, chunk_t **chunk, metalink_resource_list_t **resource, 
	int *header)
{
	*chunk = find_free_chunk(file);

	if (file->size >= 0) {
		*header = 0;
		if (!*chunk || !file->usable_res_top) {
			/* all chunks have an url assigned or there aren't anymore urls */
			*resource = NULL;
			*chunk = NULL;
			return NULL;
		}
	}
	else if (file->header) {
		/* already downloading the header */
		*resource = NULL;
		*chunk = NULL;
		*header = 0;
		return NULL;
	}
	else {
		*header = 1;
		file->header = 1;
	}

	if (!(*resource = find_free_resource(file)))
		return NULL;

	if (*chunk) {
		(*chunk)->used_res = *resource;
		(*chunk)->pos = 0;
	}

	return (*resource)->uri;
}

void set_metalink_file_length(metalink_file_list_t *file, long size)
{
	if (file)
		file->size = size;
}

void free_metalink_file(metalink_file_list_t *file)
{
	if (!file)
		return;

	free_chunks(file);
	free_resources(file);

	free(file);
}


/* metalink_list_t */

void push_metalink(metalink_t *metalink)
{
	metalink_list_t *elem;

	if (!metalink)
		return;

	elem = calloc(1, sizeof(metalink_list_t));
	elem->metalink = metalink;

	if (metalink_bottom)
		metalink_bottom->next = elem;
	metalink_bottom = elem;

	if (!metalink_top)
		metalink_top = elem;
}

void free_metalinks(void)
{
	metalink_list_t *tmp;
	while (metalink_top) {
		tmp = metalink_top;
		metalink_delete(metalink_top->metalink);
		metalink_top = metalink_top->next;
		free(tmp);
	}
}

void add_new_metalink(char *filename, int level)
{
	metalink_error_t err;
	metalink_t* metalink;
	metalink_file_t** files;
	metalink_resource_t** resources;
	metalink_file_list_t* new_file;

	err = metalink_parse_file(filename, &metalink);

	if (err) {
		if (is_printf(MERR))
			printf("ERROR: code=%d\n", err);
		return;
	}

	push_metalink(metalink);

	for (files = metalink->files; *files; files++) {
		/* resources not present */
		if (!(*files)->resources[0])
			continue;

		/* different OS */
		if (option_values.metalink_os && *(option_values.metalink_os)
			&& !strstrlower((*files)->os, option_values.metalink_os))
			continue;

		if (!(new_file = create_metalink_file(*files)))
			continue;

		for (resources = (*files)->resources; *resources; resources++) {
			if (is_valid_resource(*resources)) {
#ifdef DEBUG
				printf("%s,%d: %s\n", __FUNCTION__, __LINE__, 
					(*resources)->url ? (*resources)->url : "(null)");
#endif
				push_metalink_resource(new_file, *resources);
			}
		}

		if (new_file->usable_res_top)
			push_metalink_uri(new_file, level);
	}
}
