/*---------------------------------------------------------------------------
 * Copyright (C) 2008-2009 - Emanuele Bovisio
 *
 * This file is part of Mulk.
 *
 * Mulk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Mulk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mulk.  If not, see <http://www.gnu.org/licenses/>.
 *---------------------------------------------------------------------------*/


#include "parse.h"
#ifdef HAVE_TIDY_H
# include <tidy.h>
#elif defined(HAVE_TIDY_TIDY_H)
# include <tidy/tidy.h>
#else
# error "Don't know where to look for libtidy headers"
#endif
#include "jpg_obj.h"
#include "buffer_array.h"
#include "file_obj.h"

#define MAX_RANGE_LENGTH 200 

static size_t write_data_cb(char *ptr, size_t size, size_t nmemb, void *outstream)
{
	int nmemb_writed = 0;
	buffer_t *buffer = outstream;

	if (buffer->file_pt) {
#ifdef ENABLE_METALINK
		if (buffer->chunk) {
			size_t offset = buffer->chunk->start + buffer->chunk->pos;

			if ((buffer->chunk->pos + size * nmemb) > buffer->chunk->length)
				return 0;

			nmemb_writed = fwrite_offset(ptr, size, nmemb, offset, buffer->file_pt);
			buffer->chunk->pos += size * nmemb_writed;
		}
		else
#endif /* ENABLE_METALINK */
			nmemb_writed = fwrite(ptr, size, nmemb, buffer->file_pt);
	}

	return nmemb_writed * size;
}

static Bool filter_cb(TidyDoc tdoc, TidyReportLevel lvl, uint line, uint col, ctmbstr mssg)
{
	return 0;
}

int init_url(CURLM *cm)
{
	int i;
	url_list_t *url;
	UriUriA *uri;
#ifdef ENABLE_METALINK
	int header;
	chunk_t *chunk;
	metalink_resource_list_t *resource;
#endif
	char *str_url;
	CURL *eh;
	
	/* we need at least a free buffer */
	if (get_buffer(NULL) < 0)
		return 0;

#ifdef ENABLE_METALINK
	if (!(url = search_next_url(&uri, &chunk, &resource, &header)))
#else
	if (!(url = search_next_url(&uri)))
#endif
		return 0;

	eh = curl_easy_init();
	
#ifdef ENABLE_METALINK
	i = open_buffer(eh, url, uri, chunk, resource, header);
#else
	i = open_buffer(eh, url, uri);
#endif

	str_url = uri2string(uri);

#ifdef ENABLE_METALINK
	if (header) {
		curl_easy_setopt(eh, CURLOPT_WRITEFUNCTION, NULL);
		curl_easy_setopt(eh, CURLOPT_WRITEDATA, NULL);
		curl_easy_setopt(eh, CURLOPT_NOBODY, 1L);
	}
	else
#endif /* ENABLE_METALINK */
	{
		curl_easy_setopt(eh, CURLOPT_WRITEFUNCTION, write_data_cb);
		curl_easy_setopt(eh, CURLOPT_WRITEDATA, &(buffer_array[i]));
		curl_easy_setopt(eh, CURLOPT_NOBODY, 0L);

#ifdef ENABLE_METALINK
		if (chunk) {
		    char range[MAX_RANGE_LENGTH+1];

			range[MAX_RANGE_LENGTH] = 0;
			snprintf(range, MAX_RANGE_LENGTH, "%ld-%ld", chunk->start, 
				chunk->start + chunk->length - 1);

			curl_easy_setopt(eh, CURLOPT_RANGE, range);
		}
#endif /* ENABLE_METALINK */
	}

	curl_easy_setopt(eh, CURLOPT_HEADER, 0L);
	curl_easy_setopt(eh, CURLOPT_URL, str_url);
	curl_easy_setopt(eh, CURLOPT_PRIVATE, str_url);
	curl_easy_setopt(eh, CURLOPT_VERBOSE, 0L);
	curl_easy_setopt(eh, CURLOPT_NOPROGRESS, 1L);
	curl_easy_setopt(eh, CURLOPT_CONNECTTIMEOUT, 30L);
	curl_easy_setopt(eh, CURLOPT_LOW_SPEED_LIMIT, 100L);
	curl_easy_setopt(eh, CURLOPT_LOW_SPEED_TIME, 30L);

	curl_multi_add_handle(cm, eh);

	if (str_url)
		free(str_url);
	return 1;
}

static void parse_html(TidyNode tnod, int level, char *base_url)
{
	TidyNode child;

	for (child = tidyGetChild(tnod); child; child = tidyGetNext(child)) {
		if (tidyNodeIsA(child)) {
			TidyAttr attr = tidyAttrGetHREF(child);
			char *url = (char *) tidyAttrValue(attr);
			if (url && *url && level < option_values.depth)
				push_unique_url(base_url, url, level + 1);
		}
		else if (tidyNodeIsIMG(child)) {
			TidyAttr attr = tidyAttrGetSRC(child);
			char *url = (char *) tidyAttrValue(attr);
			if (url && *url && level < option_values.depth+1)
				push_unique_url(base_url, url, level + 1);
		}
		parse_html(child, level, base_url);
	}
}

void parse_urls(char *filename, char *url, int level)
{
	TidyDoc tdoc;
	int err;

	tdoc = tidyCreate();
	tidyOptSetBool(tdoc, TidyForceOutput, yes);
	tidyOptSetInt(tdoc, TidyWrapLen, 4096);
	tidySetReportFilter(tdoc, filter_cb);

	err = tidyParseFile(tdoc, filename);

	if (err >= 0) 
		err = tidyCleanAndRepair(tdoc);

	if (err >= 0)
		parse_html(tidyGetRoot(tdoc), level, url);

	tidyRelease(tdoc);
}

