/*---------------------------------------------------------------------------
 * Copyright (C) 2008-2009 - Emanuele Bovisio
 *
 * This file is part of Mulk.
 *
 * Mulk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Mulk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mulk.  If not, see <http://www.gnu.org/licenses/>.
 *---------------------------------------------------------------------------*/

#include "url_list.h"
#include "buffer_array.h"

url_list_t *top = NULL;
url_list_t *bottom = NULL;
url_list_t *pointer = NULL;


static url_list_t *create_new_element(void)
{
	url_list_t *elem;
	static int count = 1;

	elem = calloc(1, sizeof(url_list_t));
	elem->id = count++;

	if (bottom)
		bottom->next = elem;
	bottom = elem;

	if (!top)
		top = elem;

	if (!pointer)
		pointer = elem;

	return elem;
}

void push_uri(UriUriA *uri, int level)
{
	url_list_t *elem;

	if (!uri)
		return;

	elem = create_new_element();
	elem->level = level;
	elem->uri = uri;
}

void push_url(char *base_url, char *url, int level)
{
	UriUriA *new_uri = NULL;
	
	if (!(new_uri = create_absolute_uri(base_url, url)))
		return;

	push_uri(new_uri, level);
}

void push_unique_url(char *base_url, char *url, int level)
{
	url_list_t *elem;
	UriUriA *new_uri = NULL;
	
	if (!(new_uri = create_absolute_uri(base_url, url)))
		return;

	for (elem = top; elem; elem = elem->next)
		if (elem->uri && uriEqualsUriA(elem->uri, new_uri)) {
			uri_free(new_uri);
			return;
		}

	push_uri(new_uri, level);
}

url_list_t *get_next(url_list_t *next)
{
	return next ? next->next : top;
}

#ifdef ENABLE_METALINK
url_list_t *search_next_url(UriUriA **uri, chunk_t **chunk, metalink_resource_list_t **resource,
	int *header)
#else
url_list_t *search_next_url(UriUriA **uri)
#endif
{
	url_list_t *elem;
	int update_pointer = 1;

	for (elem = pointer; elem; elem = elem->next) {
		if (!elem->assigned) {
			if (elem->uri) {
				if (is_uri_compatible(elem->uri, -1)) {
					elem->assigned = 1;
					*uri = elem->uri;
#ifdef ENABLE_METALINK
					*resource = NULL;
					*chunk = NULL;
					*header = 0;
#endif
					if (update_pointer)
						pointer = elem->next;
					return elem;
				}
				else 
					update_pointer = 0;
			}
#ifdef ENABLE_METALINK
			else if (elem->metalink_uri) {
				*uri = find_next_url(elem->metalink_uri, chunk, resource, header);

				if (*chunk || *header) {
					if (*uri) {
						if (update_pointer)
							pointer = elem;
						return elem;
					}
					else
						update_pointer = 0;
				}
				else if (is_file_downloaded(elem->metalink_uri))
					elem->assigned = 1;
				else
					update_pointer = 0;
			}
#endif /* ENABLE_METALINK */
			else
				elem->assigned = 1;
		}
	}

	if (update_pointer)
		pointer = NULL;
	*uri = NULL;
#ifdef ENABLE_METALINK
	*chunk = NULL;
	*resource = NULL;
	*header = 0;
#endif
	return NULL;
}

/* remember to free the url */
UriUriA *pop_url(void)
{
	url_list_t *elem;
	UriUriA *ret_uri;

	if (!top)
		return NULL;

	elem = top;
	ret_uri = top->uri;

	if (top == pointer)
		pointer = pointer->next;
	top = top->next;

	if (!top)
		bottom = NULL;

#ifdef ENABLE_METALINK
	free_metalink_file(elem->metalink_uri);
#endif
	free(elem);

	return ret_uri;
}

void print_urls(char *filename)
{
	url_list_t *elem;
	FILE *file;

	if (!filename || !*filename)
		return;

   	if (!(file = fopen(filename, "w")))
		return;

	for (elem = top; elem; elem = elem->next) {
		char *uri_str = elem->uri ? uri2string(elem->uri) : NULL;

		fprintf(file, "ID:%d Depth:%d URL:\"%s\" Mime-Type:\"%s\"\n", elem->id, elem->level, 
			elem->uri ? uri_str : 
#ifdef ENABLE_METALINK
				(elem->metalink_uri ? elem->metalink_uri->file->name : "None"),
				elem->metalink_uri ? "" : elem->mimetype
#else
				"None", elem->mimetype
#endif
			);

		if (uri_str)
			free(uri_str);
	}

	fclose(file);
}

void print_csv_urls(char *filename)
{
	url_list_t *elem;
	FILE *file;

	if (!filename || !*filename)
		return;

   	if (!(file = fopen(filename, "w")))
		return;

	fprintf(file, "ID,Depth,URL,Mime-Type\n");
	for (elem = top; elem; elem = elem->next) {
		char *uri_str = elem->uri ? uri2string(elem->uri) : NULL;

		fprintf(file, "%d,%d,\"%s\",\"%s\"\n", elem->id, elem->level, 
			elem->uri ? uri_str : 
#ifdef ENABLE_METALINK
				(elem->metalink_uri ? elem->metalink_uri->file->name : "None"),
				elem->metalink_uri ? "" : elem->mimetype
#else
				"None", elem->mimetype
#endif
			);

		if (uri_str)
			free(uri_str);
	}

	fclose(file);
}

void free_urls(void)
{
	while (top)
	{
		UriUriA *uri = pop_url();
		if (uri)
			uri_free(uri);
	}
}

#ifdef ENABLE_METALINK
void push_metalink_uri(metalink_file_list_t *metalink_uri, int level)
{
	url_list_t *elem;

	if (!metalink_uri)
		return;

	elem = create_new_element();
	elem->level = level;
	elem->metalink_uri = metalink_uri;
}

void set_url_file_length(url_list_t *url, long size)
{
	if (url)
		set_metalink_file_length(url->metalink_uri, size);
}
#endif /* ENABLE_METALINK */
